from functools import reduce
from typing import Generator, List


def zip_generators(generator_list: List[Generator]):
    """
    `generator_list` 들을 모아, 단일 Generator를 생성합니다.

    Parameters
    ----------
    generator_list : List[Generator]
        [description]

    Examples
    --------
    >>> def sample_generator(f=0, t=10):
    ...     for i in range(f, t):
    ...         yield i
    
    >>> sample_generator1 = sample_generator(0, 10)
    >>> sample_generator2 = sample_generator(5, 15)
    >>> sample_generator3 = sample_generator(16, 29)
    >>> sample_generator4 = sample_generator(30, 66)
    >>> sample_generators: List[Generator] = [sample_generator1, sample_generator2, sample_generator3, sample_generator4]
    >>> list_generator = map(list, zip_generators(sample_generators))
    >>> next(list_generator)
    [0, 5, 16, 30]
    >>> next(list_generator)
    [1, 6, 17, 31]
    """
    assert len(generator_list) != 0, "List is empty."

    def _tuple_reducer(a, b):
        for _element in zip(a, b):
            if isinstance(_element[0], tuple):
                yield _element[0] + (_element[1],)
            else:
                yield (_element[0],) + (_element[1],)

    if len(generator_list) == 1:
        return map(lambda el: (el,), generator_list[0])
    else:
        return reduce(lambda a, b: _tuple_reducer(a, b), generator_list)



from typing import Optional, TypeVar

T = TypeVar("T")


def get_or_else(value_optional: Optional[T], default: T) -> T:
    """
    어떤 None이 될 수 있는 값 `value_optional`에 대해, 값이 존재하지 않는 경우 `default` 값을 리턴합니다.

    Parameters
    ----------
    value_optional : Optional[T]
        어떤 None이 될 수 있는 값
    default : T
        디폴드 값

    Returns
    -------
    T
        `value_optional` 값 혹은 디폴드 값
    """
    return default if value_optional is None else value_optional



from pathlib import Path
from typing import List, Optional, Tuple

from tensorflow.keras.preprocessing.image import DirectoryIterator, ImageDataGenerator


class FlowFromDirectory:
    """
    디렉토리에서 이미지를 읽어오는 keras의 `ImageDataGenerator`의 `flow_from_directory()` 메소드를 보조합니다.
    """

    def __init__(
        self,
        from_directory: str,
        class_directorys: List[str],
        batch_size: int,
        color_mode: str = "rgb",
        target_size: Tuple[int, int] = (256, 256),
        interpolation: str = "nearest",
        save_to_dir: Optional[str] = None,
        save_format: str = "png",
        save_prefix: str = "",
        shuffle: bool = True,
        seed: int = 42,
        class_mode: Optional[str] = None,
        subset: Optional[str] = None,
        follow_links: bool = False,
    ):
        self.from_directory = from_directory
        self.class_directorys = class_directorys
        self.batch_size = batch_size
        self.color_mode = color_mode
        self.target_size = target_size
        self.interpolation = interpolation
        self.save_to_dir = save_to_dir
        self.save_format = save_format
        self.save_prefix = save_prefix
        self.shuffle = shuffle
        self.seed = seed
        self.class_mode = class_mode
        self.subset = subset
        self.follow_links = follow_links

    def get_iterator(
        self, generator: ImageDataGenerator = ImageDataGenerator()
    ) -> DirectoryIterator:
        """
        디렉토리에서 이미지를 읽어오는 keras의 `ImageDataGenerator`의 `flow_from_directory()` 메소드를 보조합니다.

        Parameters
        ----------
        generator : ImageDataGenerator, optional
            데이터 증식를 위한 `ImageDataGenerator`를 직접 지정할 수 있습니다, by default ImageDataGenerator()

        Returns
        -------
        DirectoryIterator
            디렉토리로부터 만들어진 `ImageDataGenerator`
        """
        return generator.flow_from_directory(
            self.from_directory,
            classes=self.class_directorys,
            target_size=self.target_size,
            batch_size=self.batch_size,
            color_mode=self.color_mode,
            class_mode=self.class_mode,
            shuffle=self.shuffle,
            seed=self.seed,
            save_to_dir=self.save_to_dir,
            save_prefix=self.save_prefix,
            save_format=self.save_format,
            follow_links=self.follow_links,
            subset=self.subset,
            interpolation=self.interpolation,
        )


class ImagesFromDirectory(FlowFromDirectory):
    """
    디렉토리로부터 이미지를 읽어오는 클래스

    keras의 `ImageDataGenerator`의 `flow_from_directory()` 메소드를 보조하려는 목적입니다.
    폴더는 단일 클래스로 인식됩니다. (이미지 세그먼테이션 용도)
    """

    def __init__(
        self,
        dataset_directory: str,
        batch_size: int,
        color_mode: str = "rgb",
        target_size: Tuple[int, int] = (256, 256),
        interpolation: str = "nearest",
        save_to_dir: Optional[str] = None,
        save_format: str = "png",
        save_prefix: str = "",
        shuffle: bool = True,
        seed: int = 42,
        class_mode: Optional[str] = None,
        subset: Optional[str] = None,
        follow_links: bool = False,
    ):
        base_folder: str = str(Path(dataset_directory).parent)
        dataset_folder: str = str(Path(dataset_directory).stem)
        super().__init__(
            from_directory=base_folder,
            class_directorys=[dataset_folder],
            batch_size=batch_size,
            color_mode=color_mode,
            target_size=target_size,
            interpolation=interpolation,
            save_to_dir=save_to_dir,
            save_format=save_format,
            save_prefix=save_prefix,
            shuffle=shuffle,
            seed=seed,
            class_mode=class_mode,
            subset=subset,
            follow_links=follow_links,
        )



from pkgutil import extend_path

from .batch_transform import *
from .flow_directory import *
from .inout_generator import *
from .model_io import *
from .model_manager import *

__path__ = extend_path(__path__, "custom")
__path__ = extend_path(__path__, "utils")

__version__ = "0.2.5"



import keras.backend as K
import tensorflow as tf
from tensorflow.python.keras.metrics import MeanMetricWrapper


class BinaryClassMeanIoU(MeanMetricWrapper):
    def __init__(self, name="mean_iou", dtype=None, threshold=0.5):
        super(BinaryClassMeanIoU, self).__init__(
            binary_class_mean_iou, name, dtype=dtype, threshold=threshold
        )


def binary_class_mean_iou(y_true, y_pred, threshold=0.5):
    yt0 = y_true[:, :, :, 0]
    yp0 = K.cast(y_pred[:, :, :, 0] > threshold, "float32")
    inter = tf.math.count_nonzero(tf.logical_and(tf.equal(yt0, 1), tf.equal(yp0, 1)))
    union = tf.math.count_nonzero(tf.add(yt0, yp0))
    iou = tf.where(tf.equal(union, 0), 1.0, tf.cast(inter / union, "float32"))
    return iou



from typing import Optional

from tensorflow.keras.models import Model


def save_model(model_name: str, model: Model):
    """
    모델을 저장합니다.

    Parameters
    ----------
    model_name : str
        [description]
    model : keras.models.Model
        [description]

    Examples
    --------
    >>> model = ...   # Keras Model
    >>> save_model(model, "model_name.json")
    """
    model_json = model.to_json()
    with open(model_name, "w") as json_file:
        json_file.write(model_json)


def load_model(model_name: str, with_weights_path: Optional[str] = None) -> Model:
    """
    모델을 불러옵니다.

    Parameters
    ----------
    model_name : str
        [description]
    with_weights_path : Optional[str], optional
        [description], by default None

    Returns
    -------
    keras.models.Model
        [description]

    Examples
    --------
    >>> model = load_model("model_name.json")
    >>> model_with_weights = load_model("model_name.json", "weights/model_weight.hdf5")
    """
    from keras.models import model_from_json

    json_file = open(model_name, "r")
    loaded_model_json = json_file.read()
    json_file.close()
    model = model_from_json(loaded_model_json)
    if with_weights_path:
        model.load_weights(with_weights_path)
    return model



from typing import List, Optional, Tuple

from tensorflow.keras.losses import Loss
from tensorflow.keras.metrics import Metric
from tensorflow.keras.models import Model
from tensorflow.keras.optimizers import Optimizer


class LossDescriptor:
    """
    손실 설명을 위한 도구입니다.
    """

    def __init__(self, loss: Loss, weight: int = 1.0):
        self.loss = loss
        self.weight = weight


class ModelDescriptor:
    """
    모델 설명을 위한 도구입니다.

    - 모델에 해당되는 입력의 이름, 입력의 shape을 지정합니다.
    - 모델에 해당되는 출력의 이름, 출력의 shape을 지정합니다.
    """

    def __init__(
        self,
        inputs: List[Tuple[str, Tuple[int, int, int]]],
        outputs: List[Tuple[str, Tuple[int, int, int]]],
    ):
        self.inputs = inputs
        self.outputs = outputs

    def get_input_sizes(self) -> List[Tuple[int, int]]:
        """
        모델의 입력의 세로, 가로 크기를 반환합니다.

        Returns
        -------
        List[Tuple[int, int]]
            이미지의 세로, 가로 크기 리스트
        """
        return list(map(lambda el: (el[1][0], el[1][1]), self.inputs))

    def get_output_sizes(self) -> List[Tuple[int, int]]:
        """
        모델의 출력의 세로, 가로 크기를 반환합니다.

        Returns
        -------
        List[Tuple[int, int]]
            이미지의 세로, 가로 크기 리스트
        """
        return list(map(lambda el: (el[1][0], el[1][1]), self.outputs))


def compile_model(
    model: Model,
    model_descriptor: ModelDescriptor,
    optimizer: Optimizer,
    loss_list: List[LossDescriptor],
    metrics: List[Metric],
    sample_weight_mode=None,
    weighted_metrics=None,
    target_tensors=None,
    **kwargs
):
    """
    모델을 컴파일합니다.

    Parameters
    ----------
    model : Model
        컴파일 할 모델
    model_descriptor : ModelDescriptor
        모델 설명 도구
    optimizer : Optimizer
        옵티마이저
    loss_list : List[LossDescriptor]
        손실 설명 도구의 리스트
    metrics : List[Metric]
        메트릭 리스트
    sample_weight_mode : [type], optional
        `sample_weight_mode`, by default None
    weighted_metrics : [type], optional
        `weighted_metrics`, by default None
    target_tensors : [type], optional
        `target_tensors`, by default None
    """
    losses = dict(
        zip(
            list(map(lambda el: el[0], model_descriptor.outputs)),
            list(map(lambda el: el.loss, loss_list)),
        )
    )
    loss_weights = dict(
        zip(
            list(map(lambda el: el[0], model_descriptor.outputs)),
            list(map(lambda el: el.weight, loss_list)),
        )
    )
    model.compile(
        optimizer=optimizer,
        loss=losses,
        loss_weights=loss_weights,
        metrics=metrics,
        weighted_metrics=weighted_metrics,
        sample_weight_mode=sample_weight_mode,
        target_tensors=target_tensors,
        **kwargs
    )


class ModelHelper:
    """
    [Interface]
    """

    def __init__(self, model_descriptor: ModelDescriptor, alpha: float = 1.0):
        self.model_descriptor = model_descriptor
        self.alpha = alpha

    def get_model(self) -> Model:
        raise NotImplementedError

    def compile_model(
        self,
        model: Model,
        optimizer: Optimizer,
        loss_list: List[LossDescriptor],
        metrics: List[Metric] = [],
        model_descriptor: Optional[ModelDescriptor] = None,
        sample_weight_mode=None,
        weighted_metrics=None,
        target_tensors=None,
        **kwargs
    ) -> Model:
        _model_descriptor = model_descriptor
        if model_descriptor is None:
            _model_descriptor = self.model_descriptor
        compile_model(
            model,
            model_descriptor=_model_descriptor,
            optimizer=optimizer,
            loss_list=loss_list,
            metrics=metrics,
            sample_weight_mode=sample_weight_mode,
            weighted_metrics=weighted_metrics,
            target_tensors=target_tensors,
            **kwargs
        )
        return model



from enum import Enum
from typing import Tuple

import cv2
import numpy as np


class InterpolationEnum(Enum):
    inter_nearest = "inter_nearest"  # a nearest-neighbor interpolation
    inter_linear = "inter_linear"  # a bilinear interpolation (used by default)
    inter_area = "inter_area"  # resampling using pixel area relation. It may be a preferred method for image decimation, as it gives moire’-free results. But when the image is zoomed, it is similar to the INTER_NEAREST method.
    inter_cubic = "inter_cubic"  # a bicubic interpolation over 4×4 pixel neighborhood
    inter_lanczos4 = (
        "inter_lanczos4"  # a Lanczos interpolation over 8×8 pixel neighborhood
    )

    def get_cv2_interpolation(self):
        if self == InterpolationEnum.inter_nearest:
            return cv2.INTER_NEAREST
        elif self == InterpolationEnum.inter_linear:
            return cv2.INTER_LINEAR
        elif self == InterpolationEnum.inter_area:
            return cv2.INTER_AREA
        elif self == InterpolationEnum.inter_cubic:
            return cv2.INTER_CUBIC
        elif self == InterpolationEnum.inter_lanczos4:
            return cv2.INTER_LANCZOS4


def img_resize(
    img: np.ndarray,
    target_size: Tuple[int, int],
    interpolation: InterpolationEnum = InterpolationEnum.inter_nearest,
):
    return cv2.resize(
        img, target_size, interpolation=interpolation.get_cv2_interpolation()
    )


def gray_image_apply_clahe(
    gray_cv2_img: np.ndarray,
    clip_limit: float = 2.0,
    tile_grid_size: Tuple[int, int] = (8, 8),
) -> np.ndarray:
    """
    clahe 필터를 적용합니다. 그레이스케일 이미지에서만 동작합니다.

    Parameters
    ----------
    gray_cv2_img : np.ndarray
        그레이스케일 이미지
    clip_limit : float, optional
        Clip limit, by default 2.0
    tile_grid_size : Tuple[int, int], optional
        타일 그리드 사이즈, by default (8, 8)

    Returns
    -------
    np.ndarray
        clahe 필터를 적용한 이미지
    """
    clahe = cv2.createCLAHE(clipLimit=clip_limit, tileGridSize=tile_grid_size)
    result_img = clahe.apply(gray_cv2_img)
    return result_img


def img_to_ratio(img: np.ndarray) -> np.ndarray:
    """
    이미지를 비율로 전환합니다.

    Parameters
    ----------
    img : np.ndarray
        비율로 변환할 이미지

    Returns
    -------
    np.ndarray
        비율
    """
    return img / 255.0


def ratio_to_img(ratio_img: np.ndarray) -> np.ndarray:
    """
    비율을 이미지로 전환합니다.

    Parameters
    ----------
    img : np.ndarray
        이미지로 변환할 비율

    Returns
    -------
    np.ndarray
        이미지
    """
    return ratio_img * 255


def img_to_minmax(
    img: np.ndarray,
    threshold: float = 0.5,
    min_max: Tuple[float, float] = (0.0, 1.0),
) -> np.ndarray:
    """
    Threshold를 기준으로, 최소, 최대로 변경합니다.

    Parameters
    ----------
    img : np.ndarray
        이미지
    threshold : float, optional
        Threshold, by default 0.5
    min_max : Tuple[float, float], optional
        최소, 최대 값, by default (0.0, 1.0)

    Returns
    -------
    np.ndarray
        최소, 최대로 변환된 이미지
    """
    result_img = img.copy()
    result_img[result_img > threshold] = min_max[1]
    result_img[result_img <= threshold] = min_max[0]
    return result_img



from tensorflow.keras.callbacks import EarlyStopping, ModelCheckpoint, ReduceLROnPlateau


class EarlyStoppingAfter(EarlyStopping):
    """
    [summary]

    Parameters
    ----------
    EarlyStopping : [type]
        [description]

    Examples
    --------
    >>> apply_callbacks_after: int = 0
    >>> training_num_of_epochs: int = 200
    >>> val_freq: int = 1
    >>> early_stopping_patience: int = training_num_of_epochs // (10 * val_freq)
    >>> early_stopping: Callback = EarlyStoppingAfter(
    ...     patience=early_stopping_patience, verbose=1, after_epoch=apply_callbacks_after,
    ... )
    """

    def __init__(
        self,
        monitor="val_loss",
        min_delta=0,
        patience=0,
        verbose=0,
        mode="auto",
        baseline=None,
        restore_best_weights=False,
        after_epoch=100,
    ):
        super(EarlyStoppingAfter, self).__init__(
            monitor=monitor,
            min_delta=min_delta,
            patience=patience,
            verbose=verbose,
            mode=mode,
            baseline=baseline,
            restore_best_weights=restore_best_weights,
        )
        self.after_epoch = after_epoch

    def on_epoch_end(self, epoch, logs=None):
        if epoch >= self.after_epoch:
            super().on_epoch_end(epoch, logs)


class ModelCheckpointAfter(ModelCheckpoint):
    """
    [summary]

    Parameters
    ----------
    ModelCheckpoint : [type]
        [description]

    Examples
    --------
    >>> import time
    >>> apply_callbacks_after: int = 0
    >>> model_name: str = "unet_l4"
    >>> run_id: str = time.strftime("%Y%m%d-%H%M%S")
    >>> training_id: str = "_training__model_{}__run_{}".format(model_name, run_id)
    >>> save_weights_folder: str = os.path.join("data", "weights")
    >>> model_checkpoint: Callback = ModelCheckpointAfter(
    ...     os.path.join(
    ...         save_weights_folder,
    ...         training_id[1:]
    ...         + ".epoch_{epoch:02d}-val_loss_{val_loss:.3f}-val_mean_iou_{val_mean_iou:.3f}.hdf5",
    ...     ),
    ...     verbose=1,
    ...     # save_best_only=True,
    ...     after_epoch=apply_callbacks_after,
    ... )
    >>> callback_list: List[Callback] = [model_checkpoint]
    """

    def __init__(
        self,
        filepath,
        monitor="val_loss",
        verbose=0,
        save_best_only=False,
        save_weights_only=False,
        mode="auto",
        save_freq="epoch",
        after_epoch=100,
    ):
        super(ModelCheckpointAfter, self).__init__(
            filepath,
            monitor=monitor,
            verbose=verbose,
            save_best_only=save_best_only,
            save_weights_only=save_weights_only,
            mode=mode,
            save_freq=save_freq,
        )
        self.after_epoch = after_epoch

    def on_epoch_end(self, epoch, logs=None):
        if epoch >= self.after_epoch:
            super().on_epoch_end(epoch, logs)


class ReduceLROnPlateauAfter(ReduceLROnPlateau):
    """
    [summary]

    Parameters
    ----------
    ReduceLROnPlateau : [type]
        [description]

    Examples
    --------
    >>> apply_callbacks_after: int = 0
    >>> reduce_lr_patience: int = 10
    >>> reduce_lr_cooldown: int = 5
    >>> reduce_lr: Callback = ReduceLROnPlateauAfter(
    ...     patience=reduce_lr_patience,
    ...     cooldown=reduce_lr_cooldown,
    ...     verbose=1,
    ...     after_epoch=apply_callbacks_after,
    ...     )
    >>> callback_list: List[Callback] = [reduce_lr]
    """

    def __init__(
        self,
        monitor="val_loss",
        factor=0.1,
        patience=10,
        verbose=0,
        mode="auto",
        min_delta=1e-4,
        cooldown=0,
        min_lr=0,
        after_epoch=100,
        **kwargs
    ):
        super(ReduceLROnPlateauAfter, self).__init__(
            monitor=monitor,
            factor=factor,
            patience=patience,
            verbose=verbose,
            mode=mode,
            min_delta=min_delta,
            cooldown=cooldown,
            min_lr=min_lr,
            **kwargs
        )
        self.after_epoch = after_epoch

    def on_epoch_end(self, epoch, logs=None):
        if epoch >= self.after_epoch:
            super().on_epoch_end(epoch, logs)



from typing import Callable, Iterator, Optional

import numpy as np
import toolz

from image_keras.utils.optional import get_or_else


def transform_for_batch(
    batch_img: np.ndarray,
    each_image_transform_function: Callable[[np.ndarray], np.ndarray],
    each_transformed_image_save_function_optional: Optional[
        Callable[[int, np.ndarray], None]
    ] = None,
) -> np.ndarray:
    """
    배치 이미지를 각 배치 별로 변환하고, 저장합니다.

    Parameters
    ----------
    batch_img : np.ndarray
        변환할 배치 이미지
    each_image_transform_function : Callable[[np.ndarray], np.ndarray]
        배치 내에서 변환할 함수
    each_transformed_image_save_function_optional : Optional[Callable[[int, int, np.ndarray], None]], optional
        배치 내에서 변환 후, 이미지를 저장할 함수로, 배치 번호 및 이미지를 입력으로 받는 함수, by default None

    Returns
    -------
    np.ndarray
        배치 변환이 완료된 이미지들의 배치
    """
    img_result_list = []
    for i in range(batch_img.shape[0]):
        current_batch_img = batch_img[i, :]

        # 1. int8 정수 데이터 형태로 변환
        current_batch_img = current_batch_img.astype(np.uint8)
        # 2. 변환 함수 적용
        current_transformed_img = each_image_transform_function(current_batch_img)
        # 3. (그레이스케일 이미지의 경우,) 차원이 추가되지 않은 경우 추가
        if len(current_transformed_img.shape) == 2:
            current_transformed_img = np.reshape(
                current_transformed_img,
                (current_transformed_img.shape[0], current_transformed_img.shape[1], 1),
            )

        img_result_list.append(current_transformed_img)
        if each_transformed_image_save_function_optional:
            each_transformed_image_save_function_optional(i, current_transformed_img)
    return np.array(img_result_list)


def generate_iterator_and_transform(
    image_generator: Iterator,
    each_image_transform_function: Optional[Callable[[np.ndarray], np.ndarray]] = None,
    each_transformed_image_save_function_optional: Optional[
        Callable[[int, int, np.ndarray], None]
    ] = None,
    transform_function_for_all: Optional[Callable[[np.ndarray], np.ndarray]] = None,
) -> Iterator:
    """
    `image_from_directory`에 대해, (선택) (이미지 변환을 적용하고, 변환 결과를 저장한 다음), iterator로 출력

    배치 내 각 개별 이미지에 대해 변환 함수(`batch_transform_function`)가 적용된 후,
    배치 전체에 대한 변환 함수(`transform_function_for_all`)가 적용됩니다.

    Parameters
    ----------
    image_generator : Iterator
        파일을 가져올 `Iterator`
    each_image_transform_function : Optional[Callable[[np.ndarray], np.ndarray]], optional
        배치 내 이미지 변환 함수. 변환 함수가 지정되지 않으면, 변환 없이 그냥 내보냅니다.
    each_transformed_image_save_function_optional : Optional[Callable[[int, int, np.ndarray], None]], optional
        샘플 인덱스 번호, 배치 번호 및 이미지를 입력으로 하는 저장 함수, by default None
        `each_image_transform_function()` 메서드가 존재해야 동작합니다.
    transform_function_for_all : Optional[Callable[[np.ndarray], np.ndarray]], optional
        변환 함수. 배치 전체에 대한 변환 함수

    Yields
    -------
    Iterator
        `image_from_directory`의 generator을 가져와 변환(혹은 변환되지 않은)을 적용한 iterator
    """
    if each_image_transform_function:
        for index, batch_image in enumerate(image_generator):
            each_transformed_image_save_function_optional2 = None
            if each_transformed_image_save_function_optional:
                each_transformed_image_save_function_optional2 = toolz.curry(
                    each_transformed_image_save_function_optional
                )(index)
            batch_image = transform_for_batch(
                batch_img=batch_image,
                each_image_transform_function=get_or_else(
                    each_image_transform_function, lambda v: v
                ),
                each_transformed_image_save_function_optional=each_transformed_image_save_function_optional2,
            )
            if transform_function_for_all:
                batch_image = transform_function_for_all(batch_image)
            yield batch_image
    else:
        for batch_image in image_generator:
            if transform_function_for_all:
                batch_image = transform_function_for_all(batch_image)
            yield batch_image


def dist_generate_iterator_and_transform(
    image_generator: Iterator,
    each_image_transform_function: Optional[Callable[[np.ndarray], np.ndarray]] = None,
    each_transformed_image_save_function_optional: Optional[
        Callable[[int, int, np.ndarray], None]
    ] = None,
    transform_function_for_all: Optional[Callable[[np.ndarray], np.ndarray]] = None,
) -> Iterator:
    """
    `image_from_directory`에 대해, (선택) (이미지 변환을 적용하고, 변환 결과를 저장한 다음), iterator로 출력

    배치 내 각 개별 이미지에 대해 변환 함수(`batch_transform_function`)가 적용된 후,
    배치 전체에 대한 변환 함수(`transform_function_for_all`)가 적용됩니다.

    Parameters
    ----------
    image_generator : Iterator
        파일을 가져올 `Iterator`
    each_image_transform_function : Optional[Callable[[np.ndarray], np.ndarray]], optional
        배치 내 이미지 변환 함수. 변환 함수가 지정되지 않으면, 변환 없이 그냥 내보냅니다.
    each_transformed_image_save_function_optional : Optional[Callable[[int, int, np.ndarray], None]], optional
        샘플 인덱스 번호, 배치 번호 및 이미지를 입력으로 하는 저장 함수, by default None
        `each_image_transform_function()` 메서드가 존재해야 동작합니다.
    transform_function_for_all : Optional[Callable[[np.ndarray], np.ndarray]], optional
        변환 함수. 배치 전체에 대한 변환 함수

    Yields
    -------
    Iterator
        `image_from_directory`의 generator을 가져와 변환(혹은 변환되지 않은)을 적용한 iterator
    """
    if each_image_transform_function:
        for index, batch_image in enumerate(image_generator):
            each_transformed_image_save_function_optional2 = None
            if each_transformed_image_save_function_optional:
                each_transformed_image_save_function_optional2 = toolz.curry(
                    each_transformed_image_save_function_optional
                )(index)
            batch_image = transform_for_batch(
                batch_img=batch_image,
                each_image_transform_function=get_or_else(
                    each_image_transform_function, lambda v: v
                ),
                each_transformed_image_save_function_optional=each_transformed_image_save_function_optional2,
            )
            if transform_function_for_all:
                batch_image = transform_function_for_all(batch_image)
            yield batch_image
    else:
        for batch_image in image_generator:
            if transform_function_for_all:
                batch_image = transform_function_for_all(batch_image)
            yield batch_image



import tensorflow as tf
from tensorflow.python.keras.losses import LossFunctionWrapper, binary_crossentropy
from tensorflow.python.keras.utils import losses_utils


def my_loss_fn(y_true, y_pred):
    squared_difference = tf.square(y_true - y_pred)
    return tf.reduce_mean(squared_difference, axis=-1)  # Note the `axis=-1`


class BinaryBoundaryCrossentropy(LossFunctionWrapper):
    def __init__(
        self,
        from_logits=False,
        label_smoothing=0,
        reduction=losses_utils.ReductionV2.AUTO,
        range=1,
        max=2.0,
        name="binary_boundary_crossentropy",
    ):
        super(BinaryBoundaryCrossentropy, self).__init__(
            binary_boundary_crossentropy,
            name=name,
            reduction=reduction,
            range=range,
            max=max,
            from_logits=from_logits,
            label_smoothing=label_smoothing,
        )
        self.from_logits = from_logits
        self.range = range
        self.max = max


def binary_boundary_crossentropy(
    y_true,
    y_pred,
    from_logits=False,
    label_smoothing=0,
    range: int = 1,
    max: float = 2.0,
):
    """
    [summary]

    Parameters
    ----------
    y_true : [type]
        [description]
    y_pred : [type]
        [description]
    from_logits : bool, optional
        [description], by default False
    label_smoothing : int, optional
        [description], by default 0
    range : int, optional
        [description], by default 0
    max : float, optional
        [description], by default 1.0

    Returns
    -------
    [type]
        [description]

    Examples
    --------
    >>> from image_keras.custom.losses_binary_boundary_crossentropy import binary_boundary_crossentropy
    >>> import cv2
    >>> a = cv2.imread("tests/test_resources/a.png", cv2.IMREAD_GRAYSCALE)
    >>> a_modified = (a / 255).reshape(1, a.shape[0], a.shape[1], 1)
    >>> binary_boundary_crossentropy(a_modified, a_modified, range=1, max=2)
    """
    bce = binary_crossentropy(
        y_true=y_true,
        y_pred=y_pred,
        from_logits=from_logits,
        label_smoothing=label_smoothing,
    )

    def count_around_blocks(arr, range: int = 1):
        ones = tf.ones_like(arr)
        size = range * 2 + 1
        if range < 1:
            size = 1
        extracted = tf.image.extract_patches(
            images=ones,
            sizes=[1, size, size, 1],
            strides=[1, 1, 1, 1],
            rates=[1, 1, 1, 1],
            padding="SAME",
        )
        result = tf.reduce_sum(extracted, axis=-1)
        if range > 0:
            result -= 1
        return result

    def count_around_blocks2(arr, range: int = 1):
        size = range * 2 + 1
        if range < 1:
            size = 1
        extracted = tf.image.extract_patches(
            images=arr,
            sizes=[1, size, size, 1],
            strides=[1, 1, 1, 1],
            rates=[1, 1, 1, 1],
            padding="SAME",
        )
        e_base = extracted[:, :, :, tf.shape(extracted)[-1] // 2]
        e_base = tf.reshape(e_base, (-1, tf.shape(arr)[1], tf.shape(arr)[2], 1))
        e_base_expanded = tf.reshape(
            tf.repeat(e_base, tf.shape(extracted)[-1]),
            (-1, tf.shape(arr)[1], tf.shape(arr)[2], tf.shape(extracted)[-1]),
        )
        same_values = tf.math.equal(extracted, e_base_expanded)
        result_1 = tf.shape(extracted)[-1] - tf.cast(
            tf.math.count_nonzero(same_values, axis=-1), tf.int32
        )
        result_1 = tf.reshape(result_1, (-1, tf.shape(arr)[1], tf.shape(arr)[2], 1))
        result_1 = tf.cast(result_1, tf.float32)
        result_1 += arr
        block_counts = tf.reshape(
            count_around_blocks(arr, range), (-1, tf.shape(arr)[1], tf.shape(arr)[2], 1)
        )
        modify_result_1 = -(size ** 2 - block_counts)
        modify_result_1 = modify_result_1 * arr
        modify_result_1 = tf.cast(modify_result_1, tf.float32)
        diff_block_count = result_1 + modify_result_1
        return diff_block_count

    around_block_count = count_around_blocks(y_true, range=range)
    around_block_count = tf.reshape(
        around_block_count, (-1, tf.shape(y_true)[1], tf.shape(y_true)[2], 1)
    )
    around_block_count = tf.cast(around_block_count, tf.float32)

    diff_block_count = count_around_blocks2(y_true, range=range)
    diff_block_count = tf.reshape(
        diff_block_count, (-1, tf.shape(y_true)[1], tf.shape(y_true)[2], 1)
    )
    diff_block_count = tf.cast(diff_block_count, tf.float32)

    diff_ratio = diff_block_count / around_block_count
    diff_ratio = 1.0 + tf.cast(tf.math.maximum(max - 1.0, 0), tf.float32) * diff_ratio
    diff_ratio = tf.reshape(diff_ratio, (-1, tf.shape(y_true)[1], tf.shape(y_true)[2]))

    return bce * diff_ratio



import os
from abc import ABC, ABCMeta, abstractmethod
from itertools import tee
from typing import Callable, Generator, List, Optional, Tuple

import cv2
import numpy as np
import toolz
from tensorflow.keras.preprocessing.image import ImageDataGenerator

from image_keras.batch_transform import (
    dist_generate_iterator_and_transform,
    generate_iterator_and_transform,
)
from image_keras.flow_directory import FlowFromDirectory
from image_keras.utils.generator import zip_generators
from image_keras.utils.image_transform import InterpolationEnum, img_resize


def save_batch_transformed_img(
    target_folder: str,
    prefix: str,
    index_num: int,
    batch_num: int,
    image: np.ndarray,
) -> None:
    """
    배치 변환된 이미지를 저장합니다.

    Parameters
    ----------
    target_folder : str
        타겟 폴더
    prefix : str
        파일의 맨 앞에 붙을 prefix
    index_num : int
        파일의 인덱스 번호
    batch_num : int
        파일의 배치 번호
    image : np.ndarray
        이미지
    """
    # 이름
    img_name = "{}img_transformed_{:04d}_{:02d}.png".format(
        prefix, index_num, batch_num
    )
    img_fullpath = os.path.join(target_folder, img_name)

    # 저장
    cv2.imwrite(img_fullpath, image)


class FlowManager:
    def __init__(
        self,
        flow_from_directory: FlowFromDirectory,
        resize_to: Tuple[int, int],
        resize_interpolation: InterpolationEnum = InterpolationEnum.inter_nearest,
        image_transform_function: Optional[Callable[[np.ndarray], np.ndarray]] = None,
        each_transformed_image_save_function_optional: Optional[
            Callable[[int, int, np.ndarray], None]
        ] = None,
        transform_function_for_all: Optional[Callable[[np.ndarray], np.ndarray]] = None,
        image_data_generator: ImageDataGenerator = ImageDataGenerator(),
    ):
        """
        디렉토리에서 파일을 불러오는 매니저입니다.

        Parameters
        ----------
        flow_from_directory : FlowFromDirectory
            디렉토리부터 이미지를 읽어올 FlowFromDirectory를 지정합니다.
        resize_to: Tuple[int, int]
            이미지를 리사이즈 할 크기를 지정합니다. (세로, 가로)
        resize_interpolation: InterpolationEnum
            Interpolation 정책을 설정합니다. by default InterpolationEnum.inter_nearest
        image_data_generator : ImageDataGenerator
            ImageDataGenerator, by default ImageDataGenerator()
        image_transform_function : Optional[Callable[[np.ndarray], np.ndarray]], optional
            배치 내 이미지 변환 함수. 변환 함수가 지정되지 않으면, 변환 없이 그냥 내보냅니다., by default None
        each_transformed_image_save_function_optional : Optional[Callable[[int, int, np.ndarray], None]], optional
            샘플 인덱스 번호, 배치 번호 및 이미지를 입력으로 하는 저장 함수, by default None
        transform_function_for_all : Optional[Callable[[np.ndarray], np.ndarray]], optional
            변환 함수. 배치 전체에 대한 변환 함수, by default None
        """
        self.flow_from_directory: FlowFromDirectory = flow_from_directory
        self.image_data_generator: ImageDataGenerator = image_data_generator
        self.resize_to: Tuple[int, int] = resize_to
        _image_transform_function = lambda img: img
        if image_transform_function is not None:
            _image_transform_function = image_transform_function
        image_transform_function_with_resize = toolz.compose_left(
            lambda img: img_resize(img, resize_to, resize_interpolation),
            _image_transform_function,
        )
        self.image_transform_function = image_transform_function_with_resize
        self.each_transformed_image_save_function_optional = (
            each_transformed_image_save_function_optional
        )
        self.transform_function_for_all = transform_function_for_all


class Distributor:
    def __init__(
        self,
        resize_to: Tuple[int, int],
        resize_interpolation: InterpolationEnum = InterpolationEnum.inter_nearest,
        image_transform_function: Optional[Callable[[np.ndarray], np.ndarray]] = None,
        each_transformed_image_save_function_optional: Optional[
            Callable[[int, int, np.ndarray], None]
        ] = None,
        transform_function_for_all: Optional[Callable[[np.ndarray], np.ndarray]] = None,
    ):
        self.resize_to = resize_to
        self.resize_interpolation = resize_interpolation
        self.image_transform_function = image_transform_function
        self.each_transformed_image_save_function_optional = (
            each_transformed_image_save_function_optional
        )
        self.transform_function_for_all = transform_function_for_all


class DistFlowManager:
    def __init__(
        self,
        flow_from_directory: FlowFromDirectory,
        distributors: List[Distributor],
        image_data_generator: ImageDataGenerator = ImageDataGenerator(),
    ):
        """
        디렉토리에서 파일을 불러오는 매니저입니다.

        Parameters
        ----------
        flow_from_directory : FlowFromDirectory
            디렉토리부터 이미지를 읽어올 FlowFromDirectory를 지정합니다.
        image_data_generator : ImageDataGenerator
            ImageDataGenerator, by default ImageDataGenerator()
        """
        self.flow_from_directory: FlowFromDirectory = flow_from_directory
        self.image_data_generator: ImageDataGenerator = image_data_generator
        self.distributors: List[Distributor] = distributors


class InOutGenerator(ABC, metaclass=ABCMeta):
    """
    [Interface]
    InOutGenerator 인터페이스
    """

    @property
    @abstractmethod
    def input_flows(self) -> List[FlowManager]:
        raise NotImplementedError

    @property
    @abstractmethod
    def output_flows(self) -> List[FlowManager]:
        raise NotImplementedError


class BaseInOutGenerator(InOutGenerator):
    def __init__(
        self,
        input_flows: List[FlowManager],
        output_flows: List[FlowManager] = [],
    ):
        self.__input_flows = input_flows
        self.__output_flows = output_flows
        self.__i_generator = None

    @property
    def input_flows(self) -> List[FlowManager]:
        return self.__input_flows

    @property
    def output_flows(self) -> List[FlowManager]:
        raise self.__output_flows

    def get_samples(self) -> int:
        """
        데이터의 샘플 수를 반환합니다.

        `get_generator()` 메소드를 실행 후 반환 가능합니다. 실행하지 않은 경우, 0을 반환합니다.

        Returns
        -------
        int
            데이터의 샘플 수
        """
        return self.__i_generator.samples if self.__i_generator is not None else 0

    def get_filenames(self) -> List[str]:
        """
        데이터의 파일 이름을 반환합니다.

        `get_generator()` 메소드를 실행 후 반환 가능합니다. 실행하지 않은 경우, 비어있는 리스트 []를 반환합니다.

        Returns
        -------
        List[str]
            데이터의 파일 이름 리스트
        """
        return self.__i_generator.filenames if self.__i_generator is not None else []

    def get_generator(self) -> Generator:
        """
        입력 및 출력 Generator를 생성 및 반환합니다.

        - `input_flows`만 지정된 경우, Generator[[입력 image 1], [입력 image 2], ...]가 반환됩니다.
        - 둘 다 지정된 경우, Generator[[[입력 image 1], [입력 image 2], ...], [[출력 image 1], [출력 image 2], ...]]가 반환됩니다.

        Returns
        -------
        Generator
            입출력 Generator
        """
        input_generators: List[Generator] = []
        for index, input_flow in enumerate(self.__input_flows):
            i_generator = input_flow.flow_from_directory.get_iterator(
                input_flow.image_data_generator
            )
            if index == 0:
                self.__i_generator = i_generator
            i_transformed_generator: Generator = generate_iterator_and_transform(
                image_generator=i_generator,
                each_image_transform_function=input_flow.image_transform_function,
                each_transformed_image_save_function_optional=input_flow.each_transformed_image_save_function_optional,
                transform_function_for_all=input_flow.transform_function_for_all,
            )
            input_generators.append(i_transformed_generator)
        inputs_generator = map(list, zip_generators(input_generators))

        if len(self.__output_flows) > 0:
            output_generators: List[Generator] = []
            for output_flow in self.__output_flows:
                o_generator = output_flow.flow_from_directory.get_iterator(
                    output_flow.image_data_generator
                )
                o_transformed_generator: Generator = generate_iterator_and_transform(
                    image_generator=o_generator,
                    each_image_transform_function=output_flow.image_transform_function,
                    each_transformed_image_save_function_optional=output_flow.each_transformed_image_save_function_optional,
                    transform_function_for_all=output_flow.transform_function_for_all,
                )
                output_generators.append(o_transformed_generator)
            outputs_generator = map(list, zip_generators(output_generators))

            return zip(inputs_generator, outputs_generator)
        else:
            return inputs_generator


class DistInOutGenerator(InOutGenerator):
    def __init__(
        self,
        input_flows: List[DistFlowManager],
        output_flows: List[DistFlowManager] = [],
    ):
        self.__input_flows = input_flows
        self.__output_flows = output_flows
        self.__i_generator = None

    @property
    def input_flows(self) -> List[DistFlowManager]:
        return self.__input_flows

    @property
    def output_flows(self) -> List[DistFlowManager]:
        raise self.__output_flows

    def get_samples(self) -> int:
        """
        데이터의 샘플 수를 반환합니다.

        `get_generator()` 메소드를 실행 후 반환 가능합니다. 실행하지 않은 경우, 0을 반환합니다.

        Returns
        -------
        int
            데이터의 샘플 수
        """
        return self.__i_generator.samples if self.__i_generator is not None else 0

    def get_filenames(self) -> List[str]:
        """
        데이터의 파일 이름을 반환합니다.

        `get_generator()` 메소드를 실행 후 반환 가능합니다. 실행하지 않은 경우, 비어있는 리스트 []를 반환합니다.

        Returns
        -------
        List[str]
            데이터의 파일 이름 리스트
        """
        return self.__i_generator.filenames if self.__i_generator is not None else []

    def get_generator(self) -> Generator:
        """
        입력 및 출력 Generator를 생성 및 반환합니다.

        - `input_flows`만 지정된 경우, Generator[[입력 image 1], [입력 image 2], ...]가 반환됩니다.
        - 둘 다 지정된 경우, Generator[[[입력 image 1], [입력 image 2], ...], [[출력 image 1], [출력 image 2], ...]]가 반환됩니다.

        Returns
        -------
        Generator
            입출력 Generator
        """
        input_generators: List[Generator] = []
        for index, input_flow in enumerate(self.__input_flows):
            i_generator = input_flow.flow_from_directory.get_iterator(
                input_flow.image_data_generator
            )
            if index == 0:
                self.__i_generator = i_generator
            for distributor in input_flow.distributors:
                i_generator, i_generator_copy = tee(i_generator)
                i_transformed_generator: Generator = dist_generate_iterator_and_transform(
                    image_generator=i_generator_copy,
                    each_image_transform_function=distributor.image_transform_function,
                    each_transformed_image_save_function_optional=distributor.each_transformed_image_save_function_optional,
                    transform_function_for_all=distributor.transform_function_for_all,
                )
                input_generators.append(i_transformed_generator)
        inputs_generator = map(list, zip_generators(input_generators))

        if len(self.__output_flows) > 0:
            output_generators: List[Generator] = []
            for output_flow in self.__output_flows:
                o_generator = output_flow.flow_from_directory.get_iterator(
                    output_flow.image_data_generator
                )
                for distributor in output_flow.distributors:
                    o_generator, o_generator_copy = tee(o_generator)
                    o_transformed_generator: Generator = dist_generate_iterator_and_transform(
                        image_generator=o_generator_copy,
                        each_image_transform_function=distributor.image_transform_function,
                        each_transformed_image_save_function_optional=distributor.each_transformed_image_save_function_optional,
                        transform_function_for_all=distributor.transform_function_for_all,
                    )
                    output_generators.append(o_transformed_generator)
            outputs_generator = map(list, zip_generators(output_generators))

            return zip(inputs_generator, outputs_generator)
        else:
            return inputs_generator



from typing import List, Tuple

import numpy as np


def check_grayscale_image(img: np.ndarray) -> bool:
    """
    Open CV 이미지의 그레이스케일 이미지 여부를 체크합니다.

    Parameters
    ----------
    img : np.ndarray
        cv2 이미지

    Returns
    -------
    bool
        그레이스케일 이미지 여부
    """
    if len(img.shape) < 3:
        return True
    if img.shape[2] == 1:
        return True
    b, g, r = img[:, :, 0], img[:, :, 1], img[:, :, 2]
    if (b == g).all() and (b == r).all():
        return True
    return False


def get_colors_on(img: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """
    Open CV 이미지에서 사용된 컬러 및 개수를 반환합니다.
    그레이스케일 이미지의 경우, 색상 레벨(0-255) 및 개수를 반환합니다.

    Parameters
    ----------
    img : np.ndarray
        사용된 컬러를 분석할 cv2 이미지

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        이미지에서 사용된 컬러 및 총 개수 튜플
        (array([[  0,   0,   0], ..., [230,  82, 193]], dtype=uint8), array([298, ...,  80]))
        (array([[  0], ..., [230]], dtype=uint8), array([298, ...,  80]))

    Examples
    --------
    >>> img = cv2.imread("test.png")
    >>> img.shape
    (32, 32, 3)
    >>> get_colors_from_bgr_image(img)
    (array([[  0,   0,   0],
        [ 32, 107, 219],
        [ 33, 217, 245],
        [ 54,  54, 211],
        [154, 235, 110],
        [220, 136, 136],
        [230,  82, 193]], dtype=uint8), array([298, 119, 181, 140, 150,  56,  80]))
    >>> img2 = cv2.imread("test.png", cv2.IMREAD_GRAYSCALE)
    >>> img2.shape
    (32, 32)
    >>> get_levels_from_grayscale_image(img2)
    (array([[  0], [100], [131], [132], [145], [188], [204]], dtype=uint8), array([298, 140, 119,  80,  56, 150, 181]))
    """
    color_channel = 1 if check_grayscale_image(img) else img.shape[2]
    return np.unique(img.reshape(-1, color_channel), axis=0, return_counts=True)


def get_all_colors(
    img: np.ndarray, zero_first: bool = False
) -> List[Tuple[Tuple[int, int, int], int]]:
    """
    Open CV 컬러 이미지에서 사용된 컬러 및 개수를 반환합니다.

    `zero_first`가 `True`일 경우, 검은색을 맨 앞으로 이동합니다.
    `zero_first`가 `True`이더라도, 검은색이 없다면, 맨 앞은 검은색이 아닌 다른 색상입니다.

    Parameters
    ----------
    img : np.ndarray
        사용된 컬러를 분석할 cv2 이미지(BGR or 그레이스케일)
    zero_first : bool
        검은색(`(0, 0, 0)`, 그레이스케일일 경우, `(0, )`)을 맨 앞으로 이동합니다. by defaults False.

    Returns
    -------
    List[Tuple[Tuple[int, int, int], int]]
        이미지에서 사용된 컬러(BGR) or 색상 레벨(그레이스케일) 및 총 개수 튜플의 리스트

    Examples
    --------
    >>> img = cv2.imread("test.png")
    >>> img.shape
    (32, 32, 3)
    >>> get_all_colors(img, zero_first=True)
        [((0, 0, 0), 11272),
        ((64, 255, 0), 23679),
        ((195, 187, 90), 8256),
        ((36, 153, 187), 6437),
        ((128, 128, 0), 4466),
        ((73, 243, 229), 2094),
        ((236, 159, 0), 1507),
        ((39, 91, 217), 1494),
        ((170, 55, 92), 1422),
        ((120, 164, 75), 1254),
        ((86, 34, 148), 1083),
        ((18, 209, 220), 881),
        ((97, 23, 196), 508),
        ((0, 0, 255), 376),
        ((166, 151, 20), 344),
        ((137, 24, 193), 257),
        ((235, 90, 150), 206)]
    >>> img = cv2.imread("test.png", cv2.IMREAD_GRAYSCALE)
    >>> img.shape
    (32, 32)
    >>> get_all_colors(img, zero_first=True)
        [((0, ), 11272),
        ...
        ((235, ), 206)]
    """
    black_color = 0 if check_grayscale_image(img) else (0, 0, 0)
    image_colors = get_colors_on(img)
    colors_counts_tuple = list(map(tuple, image_colors[0])), image_colors[1]
    color_count_tuples = list(zip(colors_counts_tuple[0], colors_counts_tuple[1]))
    color_count_tuples = sorted(color_count_tuples, key=lambda x: x[1], reverse=True)
    black_color_indexes = [
        i for i, v in enumerate(color_count_tuples) if v[0] == black_color
    ]
    if len(black_color_indexes) > 0 and zero_first:
        zero_value = color_count_tuples[black_color_indexes[0]]
        color_count_tuples = (
            [zero_value]
            + color_count_tuples[: black_color_indexes[0]]
            + color_count_tuples[black_color_indexes[0] + 1 :]
        )
    return color_count_tuples



from keras import backend as K
import tensorflow as tf

# Compatible with tensorflow backend


def focal_loss(gamma=2.0, alpha=0.25):
    def focal_loss_fixed(y_true, y_pred):
        pt_1 = tf.where(tf.equal(y_true, 1), y_pred, tf.ones_like(y_pred))
        pt_0 = tf.where(tf.equal(y_true, 0), y_pred, tf.zeros_like(y_pred))
        return -K.mean(alpha * K.pow(1.0 - pt_1, gamma) * K.log(pt_1)) - K.mean(
            (1 - alpha) * K.pow(pt_0, gamma) * K.log(1.0 - pt_0)
        )

    return focal_loss_fixed



import setuptools
import os
import glob
import shutil
from os.path import abspath, dirname, normpath

import toolz
from toolz import curried
import image_keras

with open("README.md", "r") as fh:
    long_description = fh.read()

here = normpath(abspath(dirname(__file__)))


class CleanCommand(setuptools.Command):
    """Custom clean command to tidy up the project root."""

    CLEAN_FILES = "./build ./dist ./*.pyc ./*.tgz ./*.egg-info".split(" ")

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        global here

        for path_spec in self.CLEAN_FILES:
            # Make paths absolute and relative to this path
            abs_paths = glob.glob(os.path.normpath(os.path.join(here, path_spec)))
            for path in [str(p) for p in abs_paths]:
                if not path.startswith(here):
                    # Die if path in CLEAN_FILES is absolute + outside this directory
                    raise ValueError("%s is not a path inside %s" % (path, here))
                print("removing %s" % os.path.relpath(path))
                shutil.rmtree(path)


setup_requires = []

requirements = [
    ("toolz", "0.10.0"),
    ("Keras", "2.4.0"),
    ("tensorflow", "2.3.1"),
    ("common-py", "0.1.4"),
]

install_requires = toolz.pipe(
    requirements,
    curried.map(
        lambda name__min_version: name__min_version[0] + ">=" + name__min_version[1]
    ),
    list,
)

setuptools.setup(
    name="image-keras",
    cmdclass={
        "clean": CleanCommand,
    },
    version=image_keras.__version__,
    author="TenKeyLess",
    author_email="blivme84@naver.com",
    description="The fundamental package needed for common purpose with Python.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tenkeyless/image-keras",
    packages=setuptools.find_packages(exclude=["tests*"]),
    install_requires=install_requires,
    setup_requires=install_requires,
    license="MIT",
    keywords="common util",
    package_data={},
    zip_safe=False,
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS",
        "Operating System :: Unix",
    ],
    python_requires=">=3.7",
)



from typing import List, Tuple

import numpy as np

from image_keras.utils.image_info import get_all_colors


def color_map_generate(
    colors: List[Tuple[int, int, int]], add_black_if_not_exist: bool = False
) -> List[Tuple[int, Tuple[int, int, int]]]:
    """
    컬러 리스트에서 [(id, 컬러)] 리스트를 생성합니다.

    단, `(0, 0, 0)`이 있을 경우, 검은색의 id는 0으로 할당됩니다.
    `add_black_if_not_exist` 옵션에 따라, 검은 색이 리스트에 없는 경우에도 0번으로 할당 가능합니다.

    Parameters
    ----------
    colors : List[Tuple[int, int, int]]
        BGR 컬러의 리스트
    add_black_if_not_exist : bool
        만약, 컬러 리스트에 black이 없다면, black을 추가합니다.

    Returns
    -------
    List[Tuple[int, Tuple[int, int, int]]]
        ID 및 BGR 컬러의 리스트
    """
    code_tuples: List[Tuple[int, Tuple[int, int, int]]] = []
    black_exists: bool = False
    if (0, 0, 0) in colors:
        colors_without_black = list(filter(lambda el: el != (0, 0, 0), colors))
        code_tuples.append((0, (0, 0, 0)))
        black_exists = True
    else:
        colors_without_black = colors
        if add_black_if_not_exist:
            code_tuples.append((0, (0, 0, 0)))
            black_exists = True
    for index, color in enumerate(colors_without_black):
        new_index = index + 1 if black_exists else index
        code_tuples.append((new_index, color))
    return code_tuples


def image_detach(color_img: np.ndarray, bin_num: int) -> np.ndarray:
    """
    컬러 이미지 `color_img`를 색상에 따른 각 인스턴스 객체로 분리합니다.

    Parameters
    ----------
    color_img : np.ndarray
        컬러 이미지
    bin_num : int
        분리할 최대 인스턴스 객체의 개수

    Returns
    -------
    np.ndarray
        `color_img`의 width, height에 `bin_num` channel인 `np.ndarray`

    Examples
    --------
    >>> import cv2
    >>> from utils.image_transform import image_detach
    >>> img = cv2.imread("../simple_image_generator/samples/label.png")
    >>> image_detach(img, 30)
    """
    color_list: List[Tuple[int, int, int]] = list(
        map(lambda v: v[0], get_all_colors(color_img, True))
    )
    id_color_list: List[Tuple[int, Tuple[int, int, int]]] = color_map_generate(
        color_list
    )
    return image_detach_with_id_color_list(color_img, id_color_list, bin_num)


def image_detach_with_id_color_list(
    color_img: np.ndarray,
    id_color_list: List[Tuple[int, Tuple[int, int, int]]],
    bin_num: int,
    mask_value: float = 1.0,
) -> np.ndarray:
    """
    컬러 이미지 `color_img`를 색상에 따른 각 인스턴스 객체로 분리합니다.


    Parameters
    ----------
    color_img : np.ndarray
        컬러 이미지
    id_color_list : List[Tuple[int, Tuple[int, int, int]]]
        ID, BGR 컬러 튜플의 리스트. `(0, 0, 0)`이 있는 경우, `(0, 0, 0)`이 맨 앞에 옵니다.
    bin_num : int
        분리할 최대 인스턴스 객체의 개수
    mask_value : float
        색상이 존재하는 이미지의 객체에 덮어 씌울 값, by default 1.

    Returns
    -------
    np.ndarray
        `color_img`의 width, height에 `bin_num` channel인 `np.ndarray`

    """
    result: np.ndarray = np.zeros((color_img.shape[0], color_img.shape[1], bin_num))
    for index, id_color in enumerate(id_color_list):
        result_image = np.zeros(
            (color_img.shape[0], color_img.shape[1], 1), dtype=np.float32
        )
        mask = np.all(color_img == id_color[1], axis=-1)
        result_image[mask] = mask_value
        result[:, :, index : index + 1] = result_image
    return result


def image_detach_with_id_color_probability_list(
    color_img: np.ndarray,
    id_color_list: List[Tuple[int, Tuple[int, int, int]]],
    bin_num: int,
    resize_by_power_of_two: int = 0,
) -> np.ndarray:
    """
    컬러 이미지 `color_img`를 색상에 따른 각 인스턴스 확률 객체로 분리합니다.

    Parameters
    ----------
    color_img : np.ndarray
        컬러 이미지
    id_color_list : List[Tuple[int, Tuple[int, int, int]]]
        ID, BGR 컬러 튜플의 리스트. `(0, 0, 0)`이 있는 경우, `(0, 0, 0)`이 맨 앞에 옵니다.
    bin_num : int
        분리할 최대 인스턴스 객체의 개수
    resize_by_power_of_two : int
        줄일 이미지의 사이즈. 2의 제곱값. 0이면 그대로.
        예를 들어, 1이면, 1/2로 크기를 줄이고, 2이면, 1/4로 크기를 줄입니다. by default 0.

    Returns
    -------
    np.ndarray
        `color_img`의 width, height에 `bin_num` channel인 `np.ndarray`

    Examples
    --------
    >>> import cv2
    >>> test_img = cv2.imread("../../../Downloads/test.png")
    >>> from utils import image_transform
    >>> test_img_colors = image_transform.get_rgb_color_cv2(test_img, exclude_black=False)
    >>> test_img_color_map = image_transform.color_map_generate(test_img_colors)
    >>> detached_image = image_transform.image_detach_with_id_color_list(test_img, test_img_color_map, 30, 1)
    >>> detached_probability_image_0 = image_transform.image_detach_with_id_color_probability_list(test_img, test_img_color_map, 30, 0)
    >>> detached_probability_image_1 = image_transform.image_detach_with_id_color_probability_list(test_img, test_img_color_map, 30, 1)
    >>> detached_probability_image_2 = image_transform.image_detach_with_id_color_probability_list(test_img, test_img_color_map, 30, 2)
    >>> detached_probability_image_3 = image_transform.image_detach_with_id_color_probability_list(test_img, test_img_color_map, 30, 3)
    """
    result: np.ndarray = image_detach_with_id_color_list(
        color_img, id_color_list, bin_num, 1.0
    )
    ratio = 2 ** resize_by_power_of_two

    result_height, result_width, _ = result.shape
    result2 = shrink3D(result, result_height // ratio, result_width // ratio, bin_num)
    result2 = np.divide(result2, ratio ** 2)
    return result2


def shrink(data: np.ndarray, rows: int, cols: int) -> np.ndarray:
    """
    2D 배열을 요약합니다.

    Parameters
    ----------
    data : np.ndarray
        요약할 2D numpy 형태의 데이터
    rows : int
        반환할 행의 수
    cols : int
        반환할 열의 수

    Returns
    -------
    np.ndarray
        요약된 2D numpy 형태의 데이터

    Examples
    --------
    >>> import numpy as np
    >>> a = np.array([[ 1, 2, 3, 4],
            [ 5 ,6, 7, 8],
            [ 9,10,11,12],
            [13,14,15,16]])
    >>> shrink(a, 2, 2)
    array([[14, 22],    # [[1+2+5+6, 3+4+7+8],
        [46, 54]])      #  [9+10+13+14, 11+12+15+16]]
    >>> shrink(a, 2, 1)
    array([[ 36],       # [[1+2+5+6+3+4+7+8],
        [100]])         #  [9+10+13+14+11+12+15+16]]
    """
    return (
        data.reshape(rows, data.shape[0] // rows, cols, data.shape[1] // cols)
        .sum(axis=1)
        .sum(axis=2)
    )


def shrink3D(data: np.ndarray, rows: int, cols: int, channels: int) -> np.ndarray:
    """
    3D 배열을 요약합니다.

    Parameters
    ----------
    data : np.ndarray
        요약할 3D numpy 형태의 데이터
    rows : int
        반환할 행의 수
    cols : int
        반환할 열의 수
    channels : int
        반환할 채널의 수

    Returns
    -------
    np.ndarray
        요약된 2D numpy 형태의 데이터

    Examples
    --------
    >>> import numpy as np
    >>> a = np.array([[[ 1, 2], [3, 4]],
            [ [5 ,6],  [7, 8]],
            [ [9,10],  [11,12]],
            [ [13,14], [15,16]]])
    >>> shrink3D(a,2,1,2)
        array([[[16, 20]],  # [[[   1+3+5+7, 2+4+6+8]],
            [[48, 52]]])    #  [[9+11+13+15, 10+12+14+16]]]
    >>> shrink3D(a,2,1,1)
        array([[[ 36]],     # [[[   1+3+5+7+2+4+6+8]],
            [[100]]])       #  [[9+11+13+15+10+12+14+16]]]
    """
    return (
        data.reshape(
            rows,
            data.shape[0] // rows,
            cols,
            data.shape[1] // cols,
            channels,
            data.shape[2] // channels,
        )
        .sum(axis=1)
        .sum(axis=2)
        .sum(axis=3)
    )
